#!/usr/bin/env python3
"""Tests for _utils decorators."""

import unittest
from ee import _utils


class EeElementStub:

  @_utils.accept_opt_prefix('opt_arg1', ('opt_arg2', 'arg2_'))
  def __init__(self, arg1=None, arg2_=None):
    self.arg1 = arg1
    self.arg2_ = arg2_

  @_utils.accept_opt_prefix('opt_arg1', ('opt_arg2', 'arg2_'))
  def test_method(self, arg1=None, arg2_=None):
    return (arg1, arg2_)


class UtilsTest(unittest.TestCase):

  def test_accept_opt_prefix_decorator_method(self):
    # Test constructors decorated with accept_opt_prefix.
    api = EeElementStub('a test arg', 'a new arg')
    self.assertEqual('a test arg', api.arg1)
    self.assertEqual('a new arg', api.arg2_)

    api = EeElementStub(arg1='a test arg', arg2_='a new arg')
    self.assertEqual('a test arg', api.arg1)
    self.assertEqual('a new arg', api.arg2_)

    # pylint: disable=unexpected-keyword-arg
    # pytype: disable=wrong-keyword-args
    api = EeElementStub(opt_arg1='a test arg', opt_arg2='a new arg')
    self.assertEqual('a test arg', api.arg1)
    self.assertEqual('a new arg', api.arg2_)
    # pylint: enable=unexpected-keyword-arg
    # pytype: enable=wrong-keyword-args

    # Test methods decorated with accept_opt_prefix.
    api = EeElementStub()
    self.assertEqual(
        ('a test arg', 'a new arg'),
        api.test_method('a test arg', 'a new arg'),
    )
    self.assertEqual(
        ('a test arg', 'a new arg'),
        api.test_method(arg1='a test arg', arg2_='a new arg'),
    )
    self.assertEqual(
        ('a test arg', 'a new arg'),
        # pylint: disable-next=unexpected-keyword-arg
        api.test_method(opt_arg1='a test arg', opt_arg2='a new arg'),
    )

    # Test functions decorated with accept_opt_prefix.
    @_utils.accept_opt_prefix('opt_arg1', ('opt_arg2', 'arg2_'))
    def test_function(arg1=None, arg2_=None):
      return (arg1, arg2_)

    self.assertEqual(
        ('a test arg', 'a new arg'),
        test_function('a test arg', 'a new arg'),
    )
    self.assertEqual(
        ('a test arg', 'a new arg'),
        test_function(arg1='a test arg', arg2_='a new arg'),
    )
    self.assertEqual(
        ('a test arg', 'a new arg'),
        # pylint: disable-next=unexpected-keyword-arg
        test_function(opt_arg1='a test arg', opt_arg2='a new arg'),
    )
    # pylint: disable=unexpected-keyword-arg
    # pytype: disable=wrong-keyword-args
    self.assertEqual(
        ('a test arg', 'an overridden arg'),
        test_function(
            opt_arg1='a test arg',
            opt_arg2='a new arg',
            arg2_='an overridden arg',
        ),
    )
    # pylint: enable=unexpected-keyword-arg
    # pytype: enable=wrong-keyword-args


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for ee.Reducer module."""

import json

import unittest
import ee
from ee import apitestcase

TO_LIST = 'Reducer.toList'


class ReducerTest(apitestcase.ApiTestCase):

  def test_simple(self):
    reducer = ee.Reducer.toList()
    self.assertEqual({'value': 'fakeValue'}, reducer.getInfo())

    reducer_func = ee.ApiFunction.lookup(TO_LIST)
    self.assertEqual(reducer_func, reducer.func)

    self.assertFalse(reducer.isVariable())
    self.assertEqual({}, reducer.args)

    result = json.loads(reducer.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'functionName': TO_LIST,
                    'arguments': {},
                }
            }
        },
    }
    self.assertEqual(expect, result)

    cast_result = json.loads(ee.Reducer(reducer).serialize())
    self.assertEqual(expect, cast_result)

  def test_no_args(self):
    message = r"missing 1 required positional argument: 'reducer'"
    with self.assertRaisesRegex(TypeError, message):
      ee.Reducer()  # pytype: disable=missing-parameter

  def test_bad_arg_literal(self):
    message = (
        r"Reducer can only be used as a cast to Reducer. Found <class 'int'>\."
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Reducer(1)  # pytype: disable=wrong-arg-types


if __name__ == '__main__':
  unittest.main()



"""Interfaces implemented by serializable objects."""


class Encodable:
  """An interface implemented by objects that can serialize themselves."""

  def encode(self, encoder):
    """Encodes the object in a format compatible with Serializer.

    Args:
      encoder: A function that can be called to encode the components of
          an object.

    Returns:
      The encoded form of the object.
    """
    raise NotImplementedError('Encodable classes must implement encode().')

  def encode_cloud_value(self, encoder):
    """Encodes the object as a ValueNode.

    Args:
      encoder: A function that can be called to encode the components of
          an object.

    Returns:
      The encoded form of the object.
    """
    raise NotImplementedError(
        'Encodable classes must implement encode_cloud_value().')


class EncodableFunction:
  """An interface implemented by functions that can serialize themselves."""

  def encode_invocation(self, encoder):
    """Encodes the function in a format compatible with Serializer.

    Args:
      encoder: A function that can be called to encode the components of
          an object.

    Returns:
      The encoded form of the function.
    """
    raise NotImplementedError(
        'EncodableFunction classes must implement encode_invocation().')

  def encode_cloud_invocation(self, encoder):
    """Encodes the function as a FunctionInvocation.

    Args:
      encoder: A function that can be called to encode the components of
          an object. Returns a reference to the encoded value.

    Returns:
      The encoded form of the function.
    """
    raise NotImplementedError(
        'EncodableFunction classes must implement encode_cloud_invocation().')



"""Decorators to handle various deprecations."""

from __future__ import annotations

import dataclasses
import datetime
import functools
import inspect
import json
from typing import Any, Callable, Dict, Optional
import urllib
import warnings

_DEPRECATED_OBJECT = 'earthengine-stac/catalog/catalog_deprecated.json'
_DEPRECATED_ASSETS_URL = f'https://storage.googleapis.com/{_DEPRECATED_OBJECT}'

# Deprecation warnings are per-asset, per-initialization.
deprecated_assets: Dict[str, DeprecatedAsset] = None


def Deprecated(message: str):
  """Returns a decorator with a given warning message."""

  def Decorator(func):
    """Emits a deprecation warning when the decorated function is called.

    Also adds the deprecation message to the function's docstring.

    Args:
      func: The function to deprecate.

    Returns:
      func: The wrapped function.
    """

    @functools.wraps(func)
    def Wrapper(*args, **kwargs):
      warnings.warn_explicit(
          '%s() is deprecated: %s' % (func.__name__, message),
          category=DeprecationWarning,
          filename=func.__code__.co_filename,
          lineno=func.__code__.co_firstlineno + 1,
      )
      return func(*args, **kwargs)

    deprecation_message = '\nDEPRECATED: ' + message
    Wrapper.__doc__ += deprecation_message
    return Wrapper

  return Decorator


def CanUseDeprecated(func):
  """Ignores deprecation warnings emitted while the decorated function runs."""

  @functools.wraps(func)
  def Wrapper(*args, **kwargs):
    with warnings.catch_warnings():
      warnings.filterwarnings('ignore', category=DeprecationWarning)
      return func(*args, **kwargs)

  return Wrapper


@dataclasses.dataclass
class DeprecatedAsset:
  """Class for keeping track of a single deprecated asset."""

  id: str
  replacement_id: Optional[str]
  removal_date: Optional[datetime.datetime]
  learn_more_url: Optional[str]

  has_warning_been_issued: bool = False

  @classmethod
  def _ParseDateString(cls, date_str: str) -> Optional[datetime.datetime]:
    try:
      # We can't use `datetime.datetime.fromisoformat` because it's behavior
      # changes by Python version.
      return datetime.datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S%z')
    except ValueError:
      return None

  @classmethod
  def FromStacLink(cls, stac_link: Dict[str, Any]) -> DeprecatedAsset:
    removal_date = stac_link.get('gee:removal_date')
    if removal_date is not None:
      removal_date = cls._ParseDateString(removal_date)
    return DeprecatedAsset(
        id=stac_link.get('title'),
        replacement_id=stac_link.get('gee:replacement_id'),
        removal_date=removal_date,
        learn_more_url=stac_link.get('gee:learn_more_url'),
    )


def WarnForDeprecatedAsset(arg_name: str) -> Callable[..., Any]:
  """Decorator to warn on usage of deprecated assets.

  Args:
    arg_name: The name of the argument to check for asset deprecation.

  Returns:
    The decorated function which checks for asset deprecation.
  """

  def Decorator(func: Callable[..., Any]):
    @functools.wraps(func)
    def Wrapper(*args, **kwargs) -> Callable[..., Any]:
      argspec = inspect.getfullargspec(func)
      index = argspec.args.index(arg_name)
      if kwargs.get(arg_name):
        asset_name_object = kwargs[arg_name]
      elif index < len(args):
        asset_name_object = args[index]
      else:
        asset_name_object = None
      asset_name = _GetStringFromObject(asset_name_object)
      if asset_name:
        asset = (deprecated_assets or {}).get(asset_name)
        if asset:
          _IssueAssetDeprecationWarning(asset)
      return func(*args, **kwargs)

    return Wrapper

  return Decorator


def InitializeDeprecatedAssets() -> None:
  # Deprecated asset functionality is not critical. A warning is enough if
  # something unexpected happens.
  try:
    _InitializeDeprecatedAssetsInternal()
  except Exception as e:  # pylint: disable=broad-except
    warnings.warn(f'Unable to initialize deprecated assets: {e}')


def _InitializeDeprecatedAssetsInternal() -> None:
  global deprecated_assets
  if deprecated_assets is not None:
    return
  _UnfilterDeprecationWarnings()

  deprecated_assets = {}
  stac = _FetchDataCatalogStac()
  for stac_link in stac.get('links', []):
    if stac_link.get('deprecated', False):
      asset = DeprecatedAsset.FromStacLink(stac_link)
      deprecated_assets[asset.id] = asset


def Reset() -> None:
  global deprecated_assets
  deprecated_assets = None


def _FetchDataCatalogStac() -> Dict[str, Any]:
  try:
    response = urllib.request.urlopen(_DEPRECATED_ASSETS_URL).read()
  except (urllib.error.HTTPError, urllib.error.URLError):
    return {}
  return json.loads(response)


def _GetStringFromObject(obj: Any) -> Optional[str]:
  if isinstance(obj, str):
    return obj
  return None


def _UnfilterDeprecationWarnings() -> None:
  """Unfilters deprecation warnings for this module."""
  warnings.filterwarnings(
      'default', category=DeprecationWarning, module=__name__
  )


def _IssueAssetDeprecationWarning(asset: DeprecatedAsset) -> None:
  """Issues a warning for a deprecated asset if one hasn't already been issued.

  Args:
    asset: The asset.
  """
  if asset.has_warning_been_issued:
    return
  asset.has_warning_been_issued = True

  warning = (
      f'\n\nAttention required for {asset.id}! You are using a deprecated'
      ' asset.\nTo ensure continued functionality, please update it'
  )
  removal_date = asset.removal_date
  if removal_date:
    # %d gives a zero-padded day. Remove the leading zero. %-d is incompatible
    # with Windows.
    formatted_date = removal_date.strftime('%B %d, %Y').replace(' 0', ' ')
    warning = warning + f' by {formatted_date}.'
  else:
    warning = warning + '.'
  if asset.learn_more_url:
    warning = warning + f'\nLearn more: {asset.learn_more_url}\n'
  warnings.warn(warning, category=DeprecationWarning)



#!/usr/bin/env python3
"""Test for the ee.element module."""

import unittest
import ee
from ee import apitestcase


class ElementTestCase(apitestcase.ApiTestCase):

  def testSet(self):
    """Verifies Element.set() keyword argument interpretation."""
    image = ee.Image(1)

    # Constant dictionary.
    def AssertProperties(expected, image):
      properties = {}
      while image.func == ee.ApiFunction.lookup('Element.set'):
        key = image.args['key']
        if not isinstance(key, str):
          key = key.encode()
        properties[key] = image.args['value']
        image = image.args['object']
      self.assertEqual(ee.Image(1), image)
      self.assertEqual(expected, properties)

    AssertProperties({'foo': 'bar'}, image.set({'foo': 'bar'}))
    AssertProperties({'foo': 'bar'}, image.set({'properties': {'foo': 'bar'}}))
    AssertProperties({'properties': 5}, image.set({'properties': 5}))
    AssertProperties({'properties': {'foo': 'bar'}, 'baz': 'quux'},
                     image.set({'properties': {'foo': 'bar'}, 'baz': 'quux'}))
    AssertProperties({'foo': 'bar', 'baz': 'quux'},
                     image.set('foo', 'bar', 'baz', 'quux'))

    # Computed dictionary.
    computed_arg = ee.ComputedObject(None, None, 'foo')

    def CheckMultiProperties(result):
      self.assertEqual(ee.ApiFunction.lookup('Element.setMulti'), result.func)
      self.assertEqual({
          'object': image,
          'properties': ee.Dictionary(computed_arg)
      }, result.args)
    CheckMultiProperties(image.set(computed_arg))
    CheckMultiProperties(image.set({'properties': computed_arg}))

  def testInitOptParams(self):
    result = ee.Element(func=None, args=None, opt_varName='test').serialize()
    self.assertIn('"0": {"argumentReference": "test"}', result)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for the ee.Kernel module."""

import json

import unittest
import ee
from ee import apitestcase


_KERNEL_JSON = {
    'result': '0',
    'values': {
        '0': {
            'functionInvocationValue': {
                'functionName': 'Kernel.square',
                'arguments': {
                    'magnitude': {'constantValue': 2},
                    'normalize': {'constantValue': True},
                    'radius': {'constantValue': 1},
                    'units': {'constantValue': 'pixels'},
                },
            }
        }
    },
}


class KernelTest(apitestcase.ApiTestCase):

  def test_square_args(self):
    kernel = ee.Kernel.square(1, 'pixels', True, 2)
    self.assertEqual({'value': 'fakeValue'}, kernel.getInfo())

    square_func = ee.ApiFunction.lookup('Kernel.square')
    self.assertEqual(square_func, kernel.func)
    self.assertFalse(kernel.isVariable())

    result = json.loads(kernel.serialize())
    self.assertEqual(_KERNEL_JSON, result)

  def test_square_kwargs(self):
    kernel = ee.Kernel.square(
        radius=1, units='pixels', normalize=True, magnitude=2
    )
    result = json.loads(kernel.serialize())
    self.assertEqual(_KERNEL_JSON, result)

  def test_cast(self):
    kernel = ee.Kernel(ee.Kernel.square(1, 'pixels', True, 2))
    result = json.loads(kernel.serialize())
    self.assertEqual(_KERNEL_JSON, result)

  @unittest.skip('Does not work on github with python <= 3.9')
  def test_no_args(self):
    message = (
        r'Kernel\.__init__\(\) missing 1 required positional argument:'
        r' \'kernel\''
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Kernel()  # pytype:disable=missing-parameter

  def test_wrong_type(self):
    message = (
        r'Kernel can only be used as a cast to Kernel\. Found <class'
        r' \'int\'>'
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Kernel(1234)  # pytype:disable=wrong-arg-types


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for the ee.Model module."""

import json

import unittest
import ee
from ee import apitestcase

MODEL = 'Model'


class ModelTest(apitestcase.ApiTestCase):

  def test_serialize(self):
    model = ee.Model.fromAiPlatformPredictor('some-project', 'some-model')
    result = json.loads(model.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'functionName': 'Model.fromAiPlatformPredictor',
                    'arguments': {
                        'projectId': {'constantValue': 'some-model'},
                        'projectName': {'constantValue': 'some-project'},
                    },
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_cast(self):
    model = ee.Model(
        ee.Model.fromAiPlatformPredictor('some-project', 'some-model')
    )
    result = json.loads(model.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'functionName': 'Model.fromAiPlatformPredictor',
                    'arguments': {
                        'projectId': {'constantValue': 'some-model'},
                        'projectName': {'constantValue': 'some-project'},
                    },
                }
            }
        },
    }
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for the ee.Clusterer module."""

import json

import unittest
import ee
from ee import apitestcase

_WEKA_COBWEB_SERIALIZED = {
    'result': '0',
    'values': {
        '0': {
            'functionInvocationValue': {
                'functionName': 'Clusterer.wekaCobweb',
                'arguments': {
                    'acuity': {'constantValue': 2},
                    'cutoff': {'constantValue': 0.01},
                    'seed': {'constantValue': 3},
                },
            }
        }
    },
}


class ClustererTest(apitestcase.ApiTestCase):

  def test_cobweb_args(self):
    clusterer = ee.Clusterer.wekaCobweb(2, 0.01, 3)
    self.assertEqual({'value': 'fakeValue'}, clusterer.getInfo())

    join_func = ee.ApiFunction.lookup('Clusterer.wekaCobweb')
    self.assertEqual(join_func, clusterer.func)
    self.assertFalse(clusterer.isVariable())

    result = json.loads(clusterer.serialize())
    self.assertEqual(_WEKA_COBWEB_SERIALIZED, result)

  def test_cobweb_kwargs(self):
    clusterer = ee.Clusterer.wekaCobweb(acuity=2, cutoff=0.01, seed=3)
    self.assertEqual({'value': 'fakeValue'}, clusterer.getInfo())

    join_func = ee.ApiFunction.lookup('Clusterer.wekaCobweb')
    self.assertEqual(join_func, clusterer.func)
    self.assertFalse(clusterer.isVariable())

    result = json.loads(clusterer.serialize())
    self.assertEqual(_WEKA_COBWEB_SERIALIZED, result)

  def test_cast(self):
    clusterer = ee.Clusterer(
        ee.Clusterer.wekaCobweb(acuity=2, cutoff=0.01, seed=3)
    )
    result = json.loads(clusterer.serialize())
    self.assertEqual(_WEKA_COBWEB_SERIALIZED, result)

  @unittest.skip('Does not work on github with python <= 3.9')
  def test_no_args(self):
    message = (
        r'Clusterer\.__init__\(\) missing 1 required positional argument:'
        r' \'clusterer\''
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Clusterer()  # pytype:disable=missing-parameter

  def test_wrong_type(self):
    message = (
        r'Clusterer can only be used as a cast to Clusterer\. Found <class'
        r' \'int\'>'
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Clusterer(1234)  # pytype:disable=wrong-arg-types


if __name__ == '__main__':
  unittest.main()
